/*
 * Example of tconv implementation with use of C language.
 * 2015.09.16 tconv_convert_test - grab invalid conversions count.
 * 2014.12.26 Initial release.
 * Written by Pawel Tatera.
 *
 * This program (file) is for free (as free beer). Feel free to experiment.
 * Just don't remove the copyright notice. Thank you...
 */

#include <stdio.h>	// prinf, f<ops>
#include <stdlib.h>	// malloc free
#include <string.h>	// memset
#include <limits.h>     // MB_LEN_MAX
#include "tconv.h"

unsigned int tconv_convert_test(tconv *, char *, long);
void usage();


int main(int argc,const char*argv[]){
  FILE *fileIP, *fileOP;
  char *bP;
  char *mP; // for malloc-s, so keep read-only
  tconv *tcP;
  long mAlloc;
  int bCode;
  unsigned int bCnt;

  static const char *const msg[]={
    "Conversion successful.", // 0
    "No tconv object.",
    "Can't tell where the file %s ends.", // 2
    "malloc() error",
    "Can not open %s\n",  // 4
    "Character conversion error. Conversion aborted..."
  };

  if(argc < 5){ usage(); return 0; }

  // Instantiate the tconv
  tcP = tconv_init(argv[1], argv[3]);
  if(tcP==NULL){
    puts(msg[1]);
    return 1;
  }

  // Open the input file and allocate memory. MB_LEN_MAX tells the longest
  // possible count of bytes needed to align a single, multi-byte-character on.
  // This is needed in case when input file holds characters 1-byte aligned but
  // the output is required to hold some characters aligned on more than byte!
  fileIP = fopen(argv[2], "r");
  if(fileIP!=NULL){
    fseek(fileIP,0,SEEK_END);
    mAlloc=ftell(fileIP);
    if(mAlloc == -1L){
      printf(msg[2], argv[2]);
      fclose(fileIP);
      return 2;
    }
    mAlloc = mAlloc * MB_LEN_MAX;
    mP = (char*) malloc(mAlloc);
    if(mP==NULL){
      puts(msg[3]);
      fclose(fileIP);
      return 3;
    }
    rewind(fileIP);
  } else {
    printf(msg[4], argv[2]);
    return 4;
  }

  // Open the output file
  fileOP = fopen(argv[4], "w");
  if(fileOP == NULL){
    free(mP); 
    fclose(fileIP);
    printf(msg[4], argv[4]);
    return 4;
  }

  memset(mP, 0, mAlloc);

  // Convert
  // In normal conditions text lines are terminated with \n, \r\n and so on.
  // This code part also allows to accept the NULL-byte as a line terminator.
  bP = mP;
  while( bCode = getc(fileIP), bCode != EOF ){
    *bP = (char) bCode;
    if(*bP==0){
      bP=mP; 
      bCnt = tconv_convert_test(tcP, mP, mAlloc);
      if(bCnt == tconvError){
        puts(msg[5]);
        bP=mP; // prevent entering code part below the loop.
        break;
      }
      // Write converted text, including the NULL-byte (++bCnt).
      fwrite(mP, 1, ++bCnt, fileOP);
    } else bP++;
  };
  // Case when there is no NULL-byte in the middle of input file and also to
  // write the last chunk of NULL-spaced content, noted above.
  if(bP!=mP){
    bCnt = tconv_convert_test(tcP, mP, mAlloc);
    if(bCnt==tconvError) puts(msg[5]);
    else {
      fwrite(mP, 1, bCnt, fileOP);
      puts(msg[0]);
    }
  }

  // Close the files
  fclose(fileIP);
  fclose(fileOP);

  // Clean up memory
  free(mP); 
  tconv_free(tcP);

  return 0;
}


unsigned int tconv_convert_test(tconv *tcP, char * mP, long mTop){
  // This is a test case that really should never happen! But if, tells the
  // library is working improperly - provides invalid conversions count, can't
  // detect own error state or there is other unknown bug - broken C++ libs...
  unsigned int bCnt = tconv_convert(tcP, mP, mP, 0);
  if(bCnt > mTop)
    if(bCnt!=tconvError){
      bCnt=tconvError;
      puts("Whoops! Invalid memory parts accessed and possibly overwritten!");
    }

  return bCnt;
}


void usage(){
  printf("Usage:\n\ttconv iEnc iFile oEnc oFile\n");
}
