// This is a test program for the tconv library.
// Feel free to experiment with it.
// See included Makefiles to learn about how to recompile it.
// Copyright 2015 TinyIT
// Support dev@tinyit.pl

#include <iostream>	// cout
#include <cstdio>	// stdout
#include <iomanip>	// setw
#include <fstream>	// ifstream
#include <cstring>	// memset

#include "tconv.h"

#define ARRAY_SIZE(X)	( sizeof(X)/sizeof(X[0]) )

using namespace std;

static unsigned short tstErCnt=0;
static unsigned short tstOkCnt=0;

static const char *encoding[] = {
#ifdef _WIN32
 "UTF8",  // 0
 "Polish_Poland.28592",  // "iso_8859-2"
 "Polish_Poland.OCP", // 2
 "",
 NULL, // 4
 "Polish_Poland.ACP", // windows-1250
#else // ! _WIN32; assume POSIX/Linux
 "pl_PL.UTF8", // 0
 "pl_PL.iso_8859-2",
 "pl_PL.850", // 2
 "",
 NULL  // 4
#endif // ! _WIN32
};

static const char ok[] = {"OK"};
static const char error[] = {"ERROR"};

static const wchar_t okL[] = {L"OK"};
static const wchar_t errorL[] = {L"ERROR"};


// stdout_rst provides a simplest possible UNDEFINED_BEHAVIOUR walkaround when
// mixing cout/wcout in one stream (stdout in this case).
//
// Microsoft implementation bug?!
// ISO C++ declarations taken from http://www.cplusplus.com
// A) FILE * freopen(const char *filename, const char *mode, FILE *stream);
// B) FILE * stdout;
// (B) declaration makes stdout an "l-value-capable", but it does not work with
// MS Visual C compiler (Microsoft Visual Studio 2008).
// Standard also says freopen can accept NULL as a filename, to force access
// mode change. This case is cought by the MS RTC as an invalid param too.
void stdout_rst(){
#ifdef _WIN32
  // freopen("CON", "w", stdout);
  // freopen("CONOUT$", "w", stdout);
  // freopen(NULL,"w",stdout);
  // none of above actually works!
#else
  // For GCC simple pointer assignment works the way one would expect!
  stdout = freopen(NULL,"w",stdout);
#endif
  // If in need of testing / signalling the freopen results, put the code here!
  // Don't forget to change the 'void' with something more useful.
}

int load_file(const char *name, char *table[], unsigned int tMax){
  ifstream f;
  char maxline[256];
  union {
    unsigned int t;
    int iRet;
  };

  f.open(name);
  if(f.is_open()){
    t=0;
    while(f.good() && (t<tMax)){
      // TODO Do I need the "\n" delim for Win32?
      f.getline(maxline, 256);
      if(f.good()){
        table[t] = new char[ f.gcount() ]; // gcount() adds the null-character.
        strcpy(table[t], maxline);
        t++;
      }
    }
    f.close();
    iRet=0;
  } else iRet=1;

  return iRet;
}


template <class TCHAR>
void table_free(TCHAR *table[], unsigned int tMax){
  unsigned int t=0;
  while(t<tMax){
    if(table[t]){ delete [] table[t]; table[t]=NULL; }
    t++;
  }
}


void wcs_len(tconv &tc, const char * const mbP[], const unsigned int iExp[]){
  unsigned int i, iRet;
  const char *sP;

  i=0;
  while(iExp[i]){
    iRet=tc.wcs_len(mbP[i]);
    if(iRet == iExp[i]){ sP=ok; tstOkCnt++; }
    else { sP=error; tstErCnt++; }
    cout << mbP[i] << endl
      << " need: " << setw(3) << iExp[i]
      << " have: " << setw(3) << iRet
      << "\t\t\t" << sP << endl;
    i++;
  }

}

void mbs_len(tconv &tc, const wchar_t * const wcP[], const unsigned int iExp[]){
  unsigned int i, iRet;
  const wchar_t *sP;

  i=0;
  stdout_rst();
  while(iExp[i]){
    iRet=tc.mbs_len(wcP[i]);
    if(iRet == iExp[i]){ sP=okL; tstOkCnt++; }
    else { sP=errorL; tstErCnt++; }
    wcout << wcP[i] << endl
      << L" need: " << setw(3) << iExp[i]
      << L" have: " << setw(3) << iRet
      << L"\t\t\t" << sP << endl;
    i++;
  }
  stdout_rst();
}


int main(){
  static const char nullStr[]={"NULL_terminated_string"};
  static const wchar_t nullStrL[]={L"NULL_terminated_string"};

  static const unsigned int wcsLenExp[] = { 5, 9, 9, 18, 18, 20, 22, 0 };
  static const unsigned int mbsLenExp[] = { 5, 18, 18, 27, 27, 29, 31, 0 };

  tconv tc;
  union {
    const char *sP;
    const wchar_t *sPL;
  };
  unsigned int iRet, i;

  // Important for all arrays is to have last entry equal NULL/0!
  const char *mbPUTF[]={0,0,0,0,0,0,0,0};
  const char *mbPISO[]={0,0,0,0,0,0,0,0};
  const wchar_t *wcP[]={0,0,0,0,0,0,0,0};
  char mbB[32];
  wchar_t wcB[32];

  // NOT TESTED: set_ienc, set_oenc, tconv_init, tconv_free

  // #1 wcs_len
  cout << endl << " *** tconv::wcs_len ***" << endl << endl;

  if(tc.wcs_len(NULL) == tc.error){ sP=ok; tstOkCnt++; }
  else { sP=error; tstErCnt++; }
  cout << "Passing NULL" << setw(30) << sP << endl;

  if( load_file("tctf_iso8859_2.txt", (char**) mbPISO, ARRAY_SIZE(mbPISO) ) ){
    cout << "Can't open file." << endl;
    return 1;
  }
  tc.set_ienc(encoding[1]);
  wcs_len(tc, (const char**) mbPISO, wcsLenExp);

  if( load_file("tctf_utf8.txt", (char**) mbPUTF, ARRAY_SIZE(mbPUTF) ) ){
    cout << "Can't open file." << endl;
    goto TCTEST_ABORT; // free mbPUTF
  }

  tc.set_ienc(encoding[0]);
  // Create and fill wcP based on mbPUTF. wcP can not be loaded from a disk file,
  // because Windows and Linux both have different sizes of wchar_t type.
  i=0;
  do {
    wcP[i] = new wchar_t[ tc.wcs_len(mbPUTF[i]) + 1 ];
    if(wcP[i]){
      sP = mbPUTF[i]; // to not alter original mbPUTF locations when converting.
      if(tc.mb_wc((wchar_t*) wcP[i], & sP)==tconv::error) break;
      // mb_wc tests could be performed now... but...
    }
    i++;
  } while(mbPUTF[i]);
  wcs_len(tc, (const char**) mbPUTF, wcsLenExp);

  // #2 mbs_len
  cout << endl << " *** tconv::mbs_len ***" << endl << endl;

  if(tc.mbs_len(NULL) == tc.error){ sP=ok; tstOkCnt++; }
  else { sP=error; tstErCnt++; }
  cout << "Passing NULL" << setw(30) << sP << endl;

  tc.set_oenc(encoding[1]);
  mbs_len(tc, (const wchar_t **) wcP, wcsLenExp);

  tc.set_oenc(encoding[0]);
  mbs_len(tc, (const wchar_t **) wcP, mbsLenExp);

  // #3 mb_wc
  cout << endl << " *** tconv::mb_wc ***" << endl << endl;

  // Passing NULL pointers
  if(tc.mb_wc((wchar_t*) wcP[0], NULL, 1) == tc.error){ sP=ok; tstOkCnt++; }
  else { sP=error; tstErCnt++; }
  cout << "Passing NULL #1" << setw(27) << sP << endl;

  if(tc.mb_wc(NULL, mbPUTF, 1) == tc.error){ sP=ok; tstOkCnt++; }
  else { sP=error; tstErCnt++; }
  cout << "Passing NULL #2" << setw(27) << sP << endl;

  if(tc.mb_wc(NULL, &mbPUTF[7], 1) == tc.error){ sP=ok; tstOkCnt++; }
  else { sP=error; tstErCnt++; }
  cout << "Passing NULL #3" << setw(27) << sP << endl;

  if(tc.mb_wc(NULL, NULL, 1) == tc.error){ sP=ok; tstOkCnt++; }
  else { sP=error; tstErCnt++; }
  cout << "Passing NULL #4" << setw(27) << sP << endl;

  // fill mbB with garbage
  i=ARRAY_SIZE(wcB)-1;
  wmemset(wcB, 61, i);
  wcB[i]= L'\0'; // put NULL-character at the end
  cout << "Filling destination location with garbage." << endl
    << "Destination location [" << (void*) wcB << "] holds: ";
  stdout_rst();
  wcout << wcB;
  stdout_rst();
  cout << endl;

  // converting NULL terminated string
  sP = nullStr;
  wmemset(wcB, 0, ARRAY_SIZE(wcB));
  i = tc.wcs_len(sP);
  iRet = tc.mb_wc(wcB, &sP, 0);
  cout << "Passing NULL-terminated string (max=0)" << endl
    << "Source pointer should be NULL, is [" << (void*) sP << "] " << endl;
  if(iRet==i && sP==NULL){ sP=ok; tstOkCnt++; }
  else { sP=error; tstErCnt++; }
  stdout_rst();
  wcout << L"Destination location [" << (void*) wcB << L"] holds: " << wcB << endl;
  stdout_rst();
  cout << setw(42) << sP << endl;

  // converting N characters from a string
  sP = mbPUTF[0]; // mbPUTF[0] holds {"abcde"}
  wmemset(wcB, 61, ARRAY_SIZE(wcB)-1); // overwriting with garbage again ('=')
  cout << "Converting 3 characters from [" << (void*) sP << "] " << mbPUTF[0]
    << endl;
  iRet = tc.mb_wc(wcB, &sP, 3);
  if(iRet==3 && sP==(mbPUTF[0]+3)){
    cout << "Source pointer moved to [" << (void*) sP << "] " << sP << endl;
    sP=ok; tstOkCnt++;
  }
  else { sP=error; tstErCnt++; }
  stdout_rst();
  wcout << L"Destination location [" << (void*) wcB << L"] holds: " << wcB << endl;
  stdout_rst();
  cout << setw(42) << sP << endl;

  // Converting 0 characters NOT POSSIBLE!
  // Why not just doing nothing?

  // #4 wc_mb
  cout << endl << " *** tconv::wc_mb ***" << endl << endl;

  // Passing NULL pointers
  if(tc.wc_mb((char*) mbPUTF[0], NULL, 1) == tc.error){ sP=ok; tstOkCnt++; }
  else { sP=error; tstErCnt++; }
  cout << "Passing NULL #1" << setw(27) << sP << endl;

  if(tc.wc_mb(NULL, wcP, 1) == tc.error){ sP=ok; tstOkCnt++; }
  else { sP=error; tstErCnt++; }
  cout << "Passing NULL #2" << setw(27) << sP << endl;

  if(tc.wc_mb(NULL, &wcP[7], 1) == tc.error){ sP=ok; tstOkCnt++; }
  else { sP=error; tstErCnt++; }
  cout << "Passing NULL #3" << setw(27) << sP << endl;

  if(tc.wc_mb(NULL, NULL, 1) == tc.error){ sP=ok; tstOkCnt++; }
  else { sP=error; tstErCnt++; }
  cout << "Passing NULL #4" << setw(27) << sP << endl;

  // fill mbB with garbage
  i=ARRAY_SIZE(mbB)-1;
  memset(mbB, 61, i);
  mbB[i]='\0'; // put NULL byte at the end
  cout << "Filling destination location with garbage." << endl
    << "Destination location [" << (void*) mbB << "] holds: " << mbB << endl;

  // converting NULL terminated string
  sPL = nullStrL;
  i = tc.mbs_len(sPL);
  iRet = tc.wc_mb(mbB, &sPL, 0);
  cout << "Passing NULL-terminated string (max=0)" << endl
    << "Source pointer should be NULL, is [" << (void*) sP << "] " << endl;
  if(iRet==i && sPL==NULL){ sP=ok; tstOkCnt++; }
  else { sP=error; tstErCnt++; }
  cout  << "Destination location [" << (void*) mbB << "] holds: "
    << mbB << endl
    << setw(42) << sP << endl;

  // converting N characters from a string
  sPL = wcP[0]; // wcP[0] holds {L"abcde"}
  memset(mbB, 61, ARRAY_SIZE(mbB)-1); // overwriting with garbage again ('=')
  stdout_rst();
  wcout << L"Converting 3 characters from [" << (void*) sPL << L"] " << wcP[0]
    << endl;
  iRet = tc.wc_mb(mbB, &sPL, 3);
  if(iRet==3 && sPL==(wcP[0]+3)){
    wcout << L"Source pointer moved to [" << (void*) sPL << L"] " << sPL
      << endl;
    sP=ok; tstOkCnt++;
  }
  else { sP=error; tstErCnt++; }
  stdout_rst();
  cout << "Destination location [" << (void*) mbB << "] holds: " << mbB << endl
    << setw(42) << sP << endl;

  // converting MORE characters than there are in a string
  // TODO


  // Converting 0 characters NOT POSSIBLE!
  // Why not just doing nothing?

  // #5 convert
  cout << endl << " *** tconv::convert ***" << endl << endl;
  // Passing NULL pointers
  if(tc.convert(NULL, mbPUTF[0], 1) == tc.error){ sP=ok; tstOkCnt++; }
  else { sP=error; tstErCnt++; }
  cout << "Passing NULL #1" << setw(27) << sP << endl;

  if(tc.convert((char*) mbPUTF[0], NULL, 1) == tc.error){ sP=ok; tstOkCnt++; }
  else { sP=error; tstErCnt++; }
  cout << "Passing NULL #2" << setw(27) << sP << endl;

  if(tc.convert(NULL, NULL, 1) == tc.error){ sP=ok; tstOkCnt++; }
  else { sP=error; tstErCnt++; }
  cout << "Passing NULL #3" << setw(27) << sP << endl;

  // converting N-chars from string; UTF > ISO
  //tc.set_ienc(encoding[0]); // see line ~164
  tc.set_oenc(encoding[1]);
  memset(mbB, 61, ARRAY_SIZE(mbB)-1); // fill with '='
  cout << "Convert 5 characters: " << encoding[0] << " > " << encoding[1]
    << endl << "Source:\t" << mbPUTF[1] << endl;
  iRet=tc.convert(mbB, mbPUTF[1], 5);
  if(iRet!=tconv::error && iRet==5){ sP=ok; tstOkCnt++; }
  else { sP=error; tstErCnt++; }
  cout << "Result:\t" << mbB << endl << setw(42) << sP << endl;

  // converting N-chars from string; ISO > UTF
  tc.set_ienc(encoding[1]);
  tc.set_oenc(encoding[0]);
  //memset(mbB, 61, ARRAY_SIZE(mbB)-1); // fill with '='
  cout << "Convert 5 characters: " << encoding[1] << " > " << encoding[0]
    << endl << "Source:\t" << mbB << endl;
  iRet=tc.convert(mbB, mbB, 5);
  if(iRet!=tconv::error && iRet==10){ sP=ok; tstOkCnt++; } // iRet==N bytes
  else { sP=error; tstErCnt++; }
  cout << "Result:\t" << mbB << endl << setw(42) << sP << endl;
  
  // converting MORE characters than there are in a string
  // TODO

  // converting null-terminated string UTF > UTF
  tc.set_ienc(encoding[0]);
  //tc.set_oenc(encoding[0]);
  i=strlen(mbPUTF[5]);
  memset(mbB, 61, ARRAY_SIZE(mbB)-1); // fill with '='
  cout << "Converting a string back to itself [" << encoding[0] << "]"
    << endl << "Source:\t" << mbPUTF[5] << " [" << i << "]" << endl;
  iRet=tc.convert(mbB,mbPUTF[5]);
  if(iRet!=tconv::error && iRet==i && strcmp(mbPUTF[5],mbB)==0)
    { sP=ok; tstOkCnt++; } // iRet==N bytes
  else { sP=error; tstErCnt++; }
  cout << "Result:\t" << mbB << " [" << iRet << "]"
    << endl << setw(42) << sP << endl;
  
  // converting null-terminated string ISO > ISO
  tc.set_ienc(encoding[1]);
  tc.set_oenc(encoding[1]);
  i=strlen(mbPISO[5]);
  memset(mbB, 61, ARRAY_SIZE(mbB)-1); // fill with '='
  cout << "Converting a string back to itself [" << encoding[1] << "]"
    << endl << "Source:\t" << mbPISO[5] << " [" << i << "]" << endl;
  iRet=tc.convert(mbB,mbPISO[5]);
  if(iRet!=tconv::error && iRet==i && strcmp(mbPISO[5],mbB)==0)
    { sP=ok; tstOkCnt++; } // iRet==N bytes
  else { sP=error; tstErCnt++; }
  cout << "Result:\t" << mbB << " [" << iRet << "]"
    << endl << setw(42) << sP << endl;

#ifdef _WIN32
//  Put here anything for Windows OS.
//  cout << "Windows OS specific tests" << endl;


#endif // _WIN32

TCTEST_ABORT:
  table_free(mbPUTF, ARRAY_SIZE(mbPUTF));
  table_free(mbPISO, ARRAY_SIZE(mbPISO));
  table_free(wcP, ARRAY_SIZE(wcP));

  cout << "Tests  OK:     " << tstOkCnt << endl
       << "       Failed: " << tstErCnt << endl
       << "       Total:  " << tstOkCnt+tstErCnt << endl
       << "Done." << endl << endl
       <<
"Please, ignore any junk bytes in the output. This is because it is a simple"
       << endl <<
"test tool, that does not change character encoding support of the terminal."
       << endl <<
"It still lacks some tests and more advanced features. If in need of support,"
       << endl <<
"please, write to dev@tinyit.pl"
       << endl;


  return 0;
}

